
---------------
--Obejct Setup
---------------

local aib = AceItemBidLocals
	
AceItemBid     = AceAddonClass:new({
	name          = aib.Title,
	version       = "0.1",
	releaseDate   = "10-25-05",
	aceCompatible = "102",
	author        = "FryGuy",
	email         = "fryguy@burntpopcorn.net",
	website       = "http://www.burntpopcorn.net",
	category      = "RAID",
	db            = AceDatabase:new("AceItemBidDB"),
	defaults      = {duration = 60, minbids = {}},
	cmd           = AceChatCmdClass:new(aib.ChatCmd, aib.ChatOpt),
})	


function AceItemBid:Initialize()
    -- Helpful closures for accessing the addon's currently loaded profile.
    self.GetOpt = function(var) return self.db:get(self.profilePath,var)    end
    self.SetOpt = function(var,val) self.db:set(self.profilePath,var,val)   end
    self.TogOpt = function(var) return self.db:toggle(self.profilePath,var) end
	self.GetMB   = function() return self.db:get("minbids") end
	self.SetMB   = function(key, val) return self.db:set({"minbids"}, key, val) end
	self.ResetMB = function() return self.db:set("minbids", {}) end
	
	if not self.GetMB() then self.ResetMB() end
end

function AceItemBid:Enable()
    self:RegisterEvent("CHAT_MSG_WHISPER", "WhisperHandler")
    
    -- set up the item table
    self.items = { }
end
	
---------------
--Chat Options
---------------
function AceItemBid:Report()
	local d 
	self.cmd:report({
		{text="Duration", val=self.GetOpt("duration")}
	})
end

-- starts a bid for an item
function AceItemBid:StartBid(arg)
	if (arg) then
		for link, color, item, name in gfind(arg, "(|c(%x+)|Hitem:(%d+:%d+:%d+:%d+)|h%[(.-)%]|h|r)") do
			if( color and item and name and link) then
				self:StartBidForItem(item, color, name, link);
			end	
		end
	end	
end

-- sets the minimum bid for an item
function AceItemBid:SetMinBid(arg)
	if (arg) then
		for link, color, item, name, minbid in gfind(arg, "(|c(%x+)|Hitem:(%d+:%d+:%d+:%d+)|h%[(.-)%]|h|r)%s+(%d+)") do
			if( color and item and name and link and minbid) then
				self.SetMB(name, minbid)
				self.cmd:report({{text=name, val=minbid}})
			end	
		end
	end	
end

function AceItemBid:SetDuration(duration)
	duration = ace.tonum(duration)
	
	if( (duration >= 10) and (duration <= 600) ) then
		self.SetOpt("duration", duration)
		self.cmd:report({{text="duration", val=duration}})
	else
		self.cmd:error(aib.InvalidDuration)
	end
end

------------------
--Main Processing
------------------

function AceItemBid:WhisperHandler(txt, sender)
	-- if we have no bids open, they don't matter
	if table.getn(self.items) == 0 then
		return
	end

	-- if text not passed, display an error
	if arg1 == nil then
		self.cmd:error(aib.WhisperError)
		return
	end
	
	-- see if we have a winner!
	_, _, id, bid = string.find(arg1, "^(%d+)%s+(%d+)$")
	if id and bid then
		id = ace.tonum(id)
		bid = ace.tonum(bid)
		if self.items[id] ~= nil then
			if self.items[id].state == "bids" then
				-- in the bidding stage.. accept everyone's bid
				table.insert(self.items[id].bids, {bidder=arg2, amount=bid})
				SendChatMessage(string.format(aib.BidAcceptedMessage, bid, self.items[id].link), "WHISPER", nil, arg2)
				return
			elseif self.items[id].state == "rebids" then
				local castbidid = nil
				for i,v in ipairs(self.items[id].bids) do
					if v.bidder == arg2 then
						castbidid = i
					end
				end
				if (not castbidid) then
					-- didn't bid
					SendChatMessage(string.format(aib.BidFailedRebidNoBid, self.items[id].link), "WHISPER", nil, arg2)
				elseif bid < self.items[id].secondbid then
					-- didn't bid high enough
					SendChatMessage(string.format(aib.BidFailedRebidNoBid, self.items[id].link), "WHISPER", nil, arg2)
				elseif self.items[id].bids[castbidid].amount >= self.items[id].highbid then
					-- doesn't need to rebid
					SendChatMessage(string.format(aib.BidFailedRebidNoBid, self.items[id].link), "WHISPER", nil, arg2)
				else
					table.insert(self.items[id].rebids, {bidder=arg2, amount=bid})
					SendChatMessage(string.format(aib.RebidAcceptedMessage, bid, self.items[id].link), "WHISPER", nil, arg2)
					self.items[id].numrebidders = self.items[id].numrebidders - 1
					if self.items[id].numrebidders == 0 then
						-- if there are no more rebids to be done, then cancel the timer
						-- and finalize the rebids
						Timex:DeleteNamedSchedule("AIBItemRebid"..id)
						self:FinalizeRebids(id)
					end
				end
				return
			end
		end
	else
		_, _, id, accept = string.find(arg1, "^(%d+)%s+(%a+)$")
		if id and accept and (accept=="yes" or accept=="no") then
			id = ace.tonum(id)
			if self.items[id] ~= nil then
				if self.items[id].state == "acceptrebid" then
					if not self.items[id].highbidder or self.items[id].highbidder ~= arg2 then -- didn't bid
						SendChatMessage(string.format(aib.BidFailedRebidAccept, self.items[id].link), "WHISPER", nil, arg2)
					else
						if accept == "yes" then
							self:AcceptRebid(id)
							Timex:DeleteNamedSchedule("AIBItemAcceptRebid"..id)
							return
						elseif accept == "no" then
							self:RejectRebid(id)
							Timex:DeleteNamedSchedule("AIBItemAcceptRebid"..id)
							return
						end
					end
				end
			end
		end
	end
	
	if arg2 ~= UnitName("player") then
		SendChatMessage(aib.InvalidBidMessage, "WHISPER", nil, arg2)
	end
end


function AceItemBid:StartBidForItem(item, color, name, link)
	local duration = self.GetOpt("duration")

	-- make sure that we know the minbid for this item
	local minbids = self.GetMB()
	if not minbids[name] then
		self.cmd:error(aib.NoMinBidForItem, link)
		return
	end
	local minbid = minbids[name]

	-- find a valid id for the item
	local id = 1
	while self.items[id] do
		id = id + 1
	end
	
	-- set up the item in the list
	self.items[id] = {
		item = item,
		color = color,
		name = name,
		link = link,
		state = "bids",
		bids = { },
		rebids = { },
		minbid = minbid
	}
	
	-- announce to raid
	SendChatMessage(string.format(aib.InitialAnnounce, id, link, duration, minbid, UnitName("player"), id), "RAID")
	
	-- start the timer
	Timex:AddNamedSchedule(
		"AIBItemBid"..id,
		duration,
		nil,
		nil,
		AceItemBid["BidDurationComplete"],
		AceItemBid,
		id
	)

	-- display the acetimer (if necessary)
	if AceTimer ~= nil then
		AceTimer:MakeBar("Bid", nil, name, duration, 1, 0, 0);
	end
end

function AceItemBid:BidDurationComplete(id)
	if not self.items[id] then
		self.cmd:error(aib.BidCompleteNoItemError, id)
		return
	end

	-- announce to raid
	SendChatMessage(string.format(aib.BidEndAnnounce, id, self.items[id].link), "RAID")

	-- display all of the bids in sorted order
	table.sort(self.items[id].bids, function(a,b) return a.amount>b.amount end)
	self.cmd:msg(aib.BidCompleteHeader, self.items[id].link)
	for i,v in ipairs(self.items[id].bids) do
		self.cmd:msg(i..":"..v.bidder..":"..v.amount)
	end
	
	-- get the first bidder and second bidder
	if table.getn(self.items[id].bids) >= 1 then
		self.items[id].highbid = self.items[id].bids[1].amount
		self.items[id].highbidder = self.items[id].bids[1].bidder
	end
	if table.getn(self.items[id].bids) >= 2 then
		self.items[id].secondbid = self.items[id].bids[2].amount
	else
		self.items[id].secondbid = self.items[id].minbid
	end
	
	if table.getn(self.items[id].bids) < 1 then
		-- no bids
		self.cmd:msg(aib.ItemNoBids, self.items[id].link)
		self.items[id] = nil
	elseif table.getn(self.items[id].bids) == 1 or (self.items[id].highbid > self.items[id].secondbid + 50) then
		-- 1 bid, or overbid. determine how much their bid should be
		if self.items[id].highbid > self.items[id].secondbid * 1.20 then
			self.cmd:msg(aib.GapRuleFixMessage, self.items[id].highbid, self.items[id].secondbid, math.ceil(self.items[id].secondbid * 1.10))
			self.items[id].highbid = self.items[id].secondbid + (math.ceil((self.items[id].highbid) * .10))
		end
		self:FinalizeItem(id, self.items[id].highbidder, self.items[id].highbid)
	elseif table.getn(self.items[id].bids) >= 2 then
		-- there is a higher bid, so let them know
		self.items[id].state = "rebids"
		
		-- send messages to everyone that
		local highbid,lowbid = self.items[id].highbid,self.items[id].secondbid
		if highbid == lowbid then
			-- if nobody had high bid, then we gotta fix that
			self.items[id].highbid = self.items[id].highbid + 1
			self.items[id].highbidder = nil
			highbid = lowbid + 1
		end
		self.items[id].numrebidders = 0
		for i,v in ipairs(self.items[id].bids) do
			if v.amount < highbid and v.amount >= lowbid then
				SendChatMessage(string.format(aib.ItemRebid, self.items[id].link, highbid, UnitName("player"), id, self.GetOpt("duration")), "WHISPER", nil, v.bidder)
				self.items[id].numrebidders = self.items[id].numrebidders + 1
			end
		end
		
		-- start a timer to timeout the rebid
		Timex:AddNamedSchedule(
			"AIBItemRebid"..id,
			self.GetOpt("duration"),
			nil,
			nil,
			AceItemBid["FinalizeRebids"],
			AceItemBid,
			id
		)
		if AceTimer ~= nil then
			AceTimer:MakeBar("Rebid", nil, self.items[id].name, duration, 1, 0, 0);
		end
		
	end
		
end

function AceItemBid:FinalizeRebids(id)
	-- display all of the rebids in sorted order
	table.sort(self.items[id].rebids, function(a,b) return a.amount>b.amount end)
	self.cmd:msg(aib.ReBidCompleteHeader, self.items[id].link)
	for i,v in ipairs(self.items[id].rebids) do
		self.cmd:msg(i..":"..v.bidder..":"..v.amount)
	end
	
	-- if there is a tied rebid, then we can't deal with this item..
	if table.getn(self.items[id].rebids) >= 2 and self.items[id].rebids[1].amount == self.items[id].rebids[2].amount then
		self.cmd:msg(aib.TiedRebids)
		self.items[id] = nil
		return
	end
	
	-- if there's no rebids, then give it to the original winner
	if table.getn(self.items[id].rebids) < 1 then
		self:FinalizeItem(id, self.items[id].highbidder, self.items[id].highbid)
		return
	end

	-- set up who got high rebid
	self.items[id].highrebid = self.items[id].rebids[1].amount
	self.items[id].highrebidder = self.items[id].rebids[1].bidder
	
	-- if there's no high bidder (i.e. first and second bid tied), then give it to the high rebid
	if not self.items[id].highbidder then
		self:FinalizeItem(id, self.items[id].highrebidder, self.items[id].highrebid)
		return
	end
	
	-- this means that there was a high bidder, so ask them if they can handle paying extra
	self.items[id].state = "acceptrebid"

	SendChatMessage(string.format(aib.ItemAcceptRebidPrompt, self.items[id].link, bid, UnitName("player"), id, UnitName("player"), id, self.GetOpt("duration")), "WHISPER", nil, self.items[id].highbidder)
	
	-- start a timer to timeout the rebid acceptance
	Timex:AddNamedSchedule(
		"AIBItemAcceptRebid"..id,
		self.GetOpt("duration"),
		nil,
		nil,
		AceItemBid["RejectRebid"],
		AceItemBid,
		id
	)
	if AceTimer ~= nil then
		AceTimer:MakeBar("Rebid", nil, self.items[id].name, duration, 1, 0, 0);
	end	

end

function AceItemBid:AcceptRebid(id)
	SendChatMessage(string.format(aib.ItemRebidAccepted, self.items[id].link), "WHISPER", nil, self.items[id].highrebidder)
	self:FinalizeItem(id, self.items[id].highbidder, self.items[id].highrebid)
end

function AceItemBid:RejectRebid(id)
	SendChatMessage(string.format(aib.ItemRebidRejected, self.items[id].link), "WHISPER", nil, self.items[id].highbidder)
	self:FinalizeItem(id, self.items[id].highrebidder, self.items[id].highrebid)
end

function AceItemBid:FinalizeItem(id, winner, bid)
	self.cmd:msg(aib.ItemWonMessageLocal, winner, self.items[id].link, bid)
	SendChatMessage(string.format(aib.ItemWonMessage, self.items[id].link, bid), "WHISPER", nil, winner)

	-- we're done with the item now
	self.items[id] = nil
end

---------------------
--Register the Addon
---------------------

AceItemBid:RegisterForLoad()

